import numpy as np
from astropy.table import Table

from xwavecal.tests.utils import FakeContext, FakeImage
from xwavecal.utils.basic_utils import median_subtract_channels_y
from xwavecal import basic


class TestBasic:
    CONTEXT = FakeContext()

    def test_gain_normalizer(self):
        image = FakeImage()
        image.data = np.ones((10, 10))
        image.set_header_val('gain', 2)
        image = basic.GainNormalizer(self.CONTEXT).do_stage(image)
        assert image.get_header_val('gain') == 1
        assert np.allclose(image.data, 2)

    def test_overscan_subtractor(self):
        image = FakeImage()
        image.data = np.ones((10, 12)).astype(float)
        image.data[:, 10:] = .5
        image.set_header_val('data_section', '[1:10,1:10]')
        image.set_header_val('overscan_section', '[1:10, 11:12]')
        image = basic.OverscanSubtractor(self.CONTEXT).do_stage(image)
        assert np.allclose(image.data[:10, :10], 0.5)

    def test_overscan_trimmer(self):
        image = FakeImage()
        image.data = np.ones((10, 12)).astype(float)
        image.data[:, 10:] = .5
        image.set_header_val('data_section', '[1:10,1:10]')
        image = basic.Trimmer(self.CONTEXT).do_stage(image)
        assert np.allclose(image.data.shape, (10, 10))


class TestBackgroundSubtract1dSpectrum:
    def test_do_stage(self):
        stage = basic.BackgroundSubtractSpectrum(FakeContext())
        image = FakeImage()
        image.data_tables = {'SPECBOX': Table({'fiber': [1, 1], 'flux': np.ones((2, 10)),
                                               'stderr': np.ones((2, 10))})}
        image = stage.do_stage(image)
        assert np.allclose(image.data_tables['SPECBOX']['flux'].data, np.zeros((2, 10)))


class TestMedianSubtractReadoutsAlongY:
    def test_do_stage(self):
        image = FakeImage()
        image.data = np.ones((4, 4))
        image.header['num_rd_channels'] = 1
        image = basic.MedianSubtractReadoutsAlongY(None).do_stage(image)
        assert np.allclose(image.data, 0)


class TestUtils:
    def test_median_subtract_channels(self):
        a = (np.arange(3) * np.ones((3, 3))).T
        assert np.allclose(median_subtract_channels_y(a, 3), 0)
        a = (np.array([1, 1, 1, 2, 2, 2]) * np.ones((6, 6))).T
        assert np.allclose(median_subtract_channels_y(a, 2), 0)



import numpy as np

import xwavecal.munge as mg


class TestRotations:
    def test_rotate(self):
        a = np.array([[1, 2], [4, 3]])
        image = type('Image', (), {'data': a, 'ivar': a})
        assert np.allclose(mg.Rot90(None).do_stage(image).data, [[2, 3], [1, 4]])

    def test_flip(self):
        a = (np.arange(2) * np.ones((2, 2))).T
        image = type('Image', (), {'data': a, 'ivar': a})
        assert np.allclose(mg.FlipVert(None).do_stage(image).data, [[1, 1], [0, 0]])
        image.data = np.arange(2) * np.ones((2, 2))
        assert np.allclose(mg.FlipHoriz(None).do_stage(image).data, [[1, 0], [1, 0]])



import pytest
import mock

from xwavecal.utils.runtime_utils import parse_args


@mock.patch('os.path.exists', return_value=True)
def test_parse_args(mock_os):
    args = parse_args(['--output-dir', 'output', '--data-paths', 'data', 'data2', '--config-file', 'config',
                       '--fpack', '--input-dir', 'input', '--frame-type', 'lampflat'])
    assert args.output_dir == 'output'
    assert args.data_paths == ['data', 'data2']
    assert args.config_file == 'config'
    assert args.input_dir == 'input'
    assert args.frame_type == 'lampflat'
    assert args.fpack is True


def test_parse_args_raises_error():
    with pytest.raises(ValueError):
        parse_args(['--config-file', 'config', '--output-dir', 'out'])


def test_parse_args_raises_on_missing_config():
    with pytest.raises(FileNotFoundError):
        parse_args(['--output-dir', 'output', '--config-file', 'config',
                    '--fpack', '--input-dir', 'input', '--frame-type', 'lampflat'])



import numpy as np


def median_subtract_channels_y(data, num_channels):
    """
    :param data: array_like
                 Input array. Must be 2D
    :param num_channels: The number of readout channels along axis=0 (rows).
    :return: ndarray
             Input array with each horizontal slice of size data.shape[0]/num_channels
             subtracted by its median.

    Examples
    --------
    >>> import numpy as np
    >>> a = (np.arange(3) * np.ones((3, 3))).T
    >>> print(a)
    array([[0.,  0.,  0.],
           [1.,  1.,  1.],
           [2.,  2.,  2.]])
    >>> print(median_subtract_channels_y(a, 3))
    array([[0.,  0.,  0.],
           [0.,  0.,  0.],
           [0.,  0.,  0.]])
    """
    reshaped = data.reshape(num_channels, data.shape[1], -1)
    medians = np.array([np.median(readout_channel) for readout_channel in reshaped])
    return np.subtract(reshaped, medians.reshape(num_channels, 1, 1)).reshape(data.shape)



import os
import tempfile
import shutil


class Translator(dict):
    def __init__(self, header_keys, type_keys):
        self.header_keys = header_keys
        self.type_keys = type_keys

    def __getitem__(self, key):
        return self.header_keys.get(key, key)


def writeto(hdu_list, filepath, fpack=False, overwrite=True, output_verify='fix+warn', quant=int(1E6)):
    if not fpack:
        hdu_list.writeto(filepath.split('.fz')[0], overwrite=overwrite, output_verify=output_verify)
    else:
        if not filepath.endswith('.fz'):
            filepath += '.fz'
        base_filename = os.path.basename(filepath).split('.fz')[0]
        with tempfile.TemporaryDirectory() as temp_directory:
            hdu_list.writeto(os.path.join(temp_directory, base_filename),
                             overwrite=overwrite, output_verify=output_verify)
            command = 'fpack -q {quantization} {temp_directory}/{basename}'
            os.system(command.format(quantization=int(quant), temp_directory=temp_directory, basename=base_filename))
            base_filename += '.fz'
            shutil.move(os.path.join(temp_directory, base_filename), filepath)



import pytest
import numpy as np

from astropy.table import Table
from astropy.io import fits
from scipy.optimize import OptimizeResult
from unittest import mock

from xwavecal.traces import TraceMaker, LoadTrace
from xwavecal.tests.utils import array_with_peaks, FakeImage, noisify_image, FakeContext
from xwavecal.utils.trace_utils import Trace, SingleTraceFitter, AllTraceFitter
from xwavecal.tests.utils import fill_image_with_traces

import logging
logger = logging.getLogger(__name__)


class FakeTraceImage(FakeImage):
    """
    Image of 500x500 is recommended. Drastic changes to that dimension may break trace testing.
    """
    def __init__(self, nx=500, ny=502, *args, **kwargs):
        super(FakeTraceImage, self).__init__(*args, **kwargs)
        self.caltype = 'TRACE'
        self.header['OBJECTS'] = 'tung&tung&none'
        self.nx = nx
        self.ny = ny
        self.bpm = np.zeros((self.ny, self.nx), dtype=np.uint8)
        self.data = np.zeros((self.ny, self.nx))
        self.fiber0_lit, self.fiber1_lit, self.fiber2_lit = False, True, True


class TestTrace:
    """
    Unit tests for the Trace class.
    """
    def test_trace_instantiates_from_num_centers(self):
        trace = Trace(num_centers_per_trace=5)
        assert trace.data.colnames == ['id', 'centers']
        assert len(trace.data['id']) == 0
        assert trace.data['centers'].shape == (0, 5)
        assert trace.data['centers'].description is not None
        assert trace.data['id'].description is not None

    def test_trace_instantiates_from_data(self):
        data = Table({'id': [1], 'centers': [[1, 2, 3]]})
        data['id'].description = 'test'
        data['centers'].description = 'test_2'
        trace = Trace(data=data)
        assert trace.filepath is None
        for name in ['id', 'centers']:
            assert name in trace.data.colnames
        assert len(trace.data.colnames) == 2
        assert len(trace.data['id']) == 1
        assert trace.data['centers'].shape == (1, 3)
        assert trace.data['centers'].description == 'test_2'
        assert trace.data['id'].description == 'test'

    def test_trace_raises_exception(self):
        with pytest.raises(Exception):
            Trace(data=None, num_centers_per_trace=0)

    def test_getting_trace_centers(self):
        trace = Trace(data={'id': [0, 1], 'centers': [[0, 1], [1, 2]]})
        assert np.allclose(trace.get_centers(0), [0, 1])

    def test_getting_trace_id(self):
        trace = Trace(data={'id': [0, 1], 'centers': [[0, 1], [1, 2]]})
        assert trace.get_id(-1) == 1
        assert trace.get_id(0) == 0

    def test_getting_num_found_traces(self):
        trace = Trace(data={'id': [0, 1], 'centers': [[0, 1], [1, 2]]})
        assert trace.num_traces_found() == 2

    def test_add_centers_to_empty(self):
        centers = np.arange(3)
        trace = Trace(data=None, num_centers_per_trace=len(centers))
        trace.add_centers(trace_centers=centers, trace_id=1)
        assert np.allclose(trace.data['centers'], [centers])
        assert np.allclose(trace.data['id'], [1])

    def test_add_centers_to_existing(self):
        centers = np.arange(3)
        trace = Trace(data={'id': [1], 'centers': [centers]})
        trace.add_centers(trace_centers=centers, trace_id=2)
        assert np.allclose(trace.data['centers'], [centers, centers])
        assert np.allclose(trace.data['id'], [1, 2])

    def test_sorting_trace_centers(self):
        centers = np.array([1, 2, 3])
        data = {'id': [1, 2, 3, 4],
                'centers': [centers, centers+5, centers-10, centers+2]}
        trace = Trace(data=data)
        trace.sort()
        assert np.allclose(trace.data['id'], np.arange(4))
        assert np.allclose(trace.data['centers'],
                           np.array([centers-10, centers, centers+2, centers+5]))

    def test_delete_centers(self):
        centers = np.array([1, 2, 3, 4])
        data = {'id': [1, 2, 3, 4],
                'centers': [centers, centers+5, centers+10, centers+11]}
        trace = Trace(data=data)
        trace.del_centers([])
        assert np.allclose(trace.data['id'], data['id'])
        assert np.allclose(trace.data['centers'], data['centers'])
        trace.del_centers(-1)
        assert np.allclose(trace.data['id'], [1, 2, 3])
        assert np.allclose(trace.data['centers'], [centers, centers+5, centers+10])
        trace.del_centers([-1, -2])
        assert np.allclose(trace.data['id'], [1])
        assert np.allclose(trace.data['centers'], [centers])

    def test_remove_duplicates(self):
        centers = np.array([1, 2, 3, 4])
        data = {'id': [0, 1, 2, 3],
                'centers': [centers, centers+1, centers+10, centers+15]}
        trace = Trace(data=data)
        trace.remove_duplicates(thresh=3)
        assert np.allclose(trace.data['id'], [0, 1, 2])
        assert np.allclose(trace.data['centers'].data, [centers+1, centers+10, centers+15])


class TestAllTraceFitter:
    @mock.patch('xwavecal.utils.trace_utils.SingleTraceFitter.update_initial_guess_to_run_through_pt')
    @mock.patch('xwavecal.utils.trace_utils.SingleTraceFitter.fit_trace')
    @mock.patch('xwavecal.utils.trace_utils.SingleTraceFitter.use_fit_as_initial_guess')
    def test_step_through_detector(self, use_last, fit, update):
        fit.return_value = np.arange(10)
        num_traces = 3
        fitter = SingleTraceFitter(extraargs={'initialize_fit_objects': False})
        trace = Trace(num_centers_per_trace=10)
        peak_xy_coordinates = list(zip(np.arange(num_traces), np.arange(num_traces)))
        trace = AllTraceFitter()._step_through_detector(trace, fitter, peak_xy_coordinates)
        for i in range(num_traces):
            assert np.allclose(trace.get_centers(i), np.arange(10))

    @mock.patch('xwavecal.utils.trace_utils.AllTraceFitter._step_through_detector')
    @mock.patch('xwavecal.utils.trace_utils.AllTraceFitter._identify_traces')
    @mock.patch('xwavecal.utils.trace_utils.SingleTraceFitter.__init__', return_value=None)
    def test_fit_traces(self, fitter, identify, step):
        step.return_value = Trace(data={'id': [0, 2, 1], 'centers': np.array([np.arange(3),
                                                                              np.arange(2, 5),
                                                                              np.arange(1, 4)])})
        trace = AllTraceFitter(min_peak_to_peak_spacing=0).fit_traces(None, None, None, None, None)
        for i in range(trace.num_traces_found()):
            assert np.allclose(trace.get_centers(i), np.arange(3)+i)
            assert np.isclose(trace.get_id(i), i)

    def test_getting_flux_down_detector(self):
        fake_trace_image = np.ones((5, 5))
        fake_trace_image[1] = 10
        fake_trace_image[3] = 10
        image_noise_estimate = np.sqrt(2)
        flux_signal = AllTraceFitter(xmin=3, xmax=5)._filtered_flux_down_detector(fake_trace_image,
                                                                                  image_noise_estimate)
        expected_slice = np.array([1, 10, 1, 10, 1])/np.sqrt(image_noise_estimate**2 + np.array([1, 10, 1, 10, 1]))
        assert np.allclose(flux_signal, expected_slice)

    @mock.patch('xwavecal.utils.trace_utils.AllTraceFitter._filtered_flux_down_detector')
    def test_identify_traces(self, flux):
        min_snr = 2
        peak_centers = np.array([15, 40, 60, 80])
        flux.return_value = array_with_peaks(x=np.arange(100).astype(float), centroids=peak_centers,
                                             amplitudes=[10, 6, min_snr+0.1, min_snr-0.1], stds=[3, 3, 3, 3])
        fitter = AllTraceFitter(min_snr=min_snr, min_peak_to_peak_spacing=5, xmin=0, xmax=2)
        peak_xy_coordinates = fitter._identify_traces(None, None)
        expected_peak_xy_coordinates = list(zip(np.ones_like(peak_centers[:-1]), peak_centers[:-1]))
        assert np.allclose(peak_xy_coordinates, expected_peak_xy_coordinates, atol=3)


class TestSingleTraceFitter:
    def test_class_attributes(self):
        fitter = SingleTraceFitter(extraargs={'initialize_fit_objects': False})
        assert fitter.second_order_coefficient_guess is None
        assert fitter.image_data is None
        assert fitter.filtered_image_data is None
        assert fitter.initial_guess_next_fit is None
        assert fitter.x is None
        assert fitter.x_norm is None
        assert fitter.design_matrix is None

    def test_default_class_attributes(self):
        fitter = SingleTraceFitter(extraargs={'initialize_fit_objects': False})
        assert fitter.poly_fit_order == 2
        assert fitter.coefficients == []

    def test_fit_initilization(self):
        """
        tests that SingleTraceFitter calls _initialize_fit_objects correctly upon
        instantiation of the class.
        """
        poly_fit_order = 2
        fitter = SingleTraceFitter(image_data=np.zeros((2, 2)),
                                   poly_fit_order=poly_fit_order,
                                   second_order_coefficient_guess=90)
        assert np.allclose(fitter.x_norm, np.array([-1, 1]))
        assert np.allclose(fitter.x, np.arange(2))
        assert fitter.filtered_image_data is not None
        assert fitter.design_matrix.shape == (poly_fit_order+1, 2)

    def test_generating_initial_guess(self):
        fitter = SingleTraceFitter(image_data=np.zeros((2, 2)),
                                   poly_fit_order=2,
                                   second_order_coefficient_guess=90)
        assert np.allclose(fitter.initial_guess_next_fit, np.array([0, 0, 90]))

    def test_generating_initial_guess_raises_error(self):
        with pytest.raises(Exception):
            SingleTraceFitter(image_data=np.zeros((2, 2)),
                              poly_fit_order=2,
                              second_order_coefficient_guess=None)

    def test_changing_initial_guesses(self):
        coefficients = [np.array([0, 0])]
        fitter = SingleTraceFitter(extraargs={'initialize_fit_objects': False,
                                              'coefficients': coefficients})
        fitter.use_fit_as_initial_guess(-1)
        assert np.allclose(fitter.initial_guess_next_fit, coefficients[-1])
        fitter.initial_guess_next_fit += 1
        assert not np.allclose(fitter.initial_guess_next_fit, coefficients[-1])

    @staticmethod
    def _generate_fitter():
        design_matrix = np.ones((2, 5))
        design_matrix[1] = np.linspace(-1, 1, 5)
        offset, linear_coefficient = 1, 0.5
        x = np.arange(design_matrix.shape[1])
        fitter = SingleTraceFitter(extraargs={'initialize_fit_objects': False,
                                              'design_matrix': design_matrix,
                                              'x': x})
        single_trace_coefficients = np.array([offset, linear_coefficient])
        fitter.initial_guess_next_fit = single_trace_coefficients
        a_line = np.linspace(offset - linear_coefficient, offset + linear_coefficient, 5)

        return fitter, a_line

    def test_centers_from_coefficients(self):
        fitter, a_line = self._generate_fitter()
        assert np.allclose(fitter._centers_from_coefficients(fitter.initial_guess_next_fit), a_line)

    def test_update_initial_guess_to_run_through_pt(self):
        fitter, a_line = self._generate_fitter()
        fitter.update_initial_guess_to_run_through_pt(xy=(1, 400))
        shifted_line = a_line - a_line[1] + 400
        assert np.allclose(fitter._centers_from_coefficients(fitter.initial_guess_next_fit), shifted_line)

    def test_flux_across_trace(self):
        x = np.arange(5)
        fake_data = np.zeros((9, len(x)))
        fake_data[3] += 1
        trace_centers = np.ones(len(x))*3
        filtered_fake_data = SingleTraceFitter._prefilter_image_data(fake_data)
        fitter = SingleTraceFitter(extraargs={'initialize_fit_objects': False,
                                              'x': x,
                                              'filtered_image_data': filtered_fake_data})
        expected_flux = len(x)
        flux = fitter._flux_across_trace(trace_centers)
        assert np.isclose(flux, expected_flux)

    def test_trace_merit_function_returns_negative_flux(self):
        x = np.arange(5)
        fake_data = np.zeros((9, len(x)))
        fake_data[3] += 1
        coefficients_for_line = np.array([3, 0])
        design_matrix = np.array([np.ones(len(x)),
                                  np.linspace(-1, 1, len(x))])
        filtered_fake_data = SingleTraceFitter._prefilter_image_data(fake_data)
        fitter = SingleTraceFitter(extraargs={'initialize_fit_objects': False,
                                              'x': x,
                                              'design_matrix': design_matrix,
                                              'filtered_image_data': filtered_fake_data})
        negative_flux = (-1)*len(x)
        merit = fitter._trace_merit_function(single_trace_coefficients=coefficients_for_line,
                                             cls=fitter)
        assert np.isclose(merit, negative_flux)

    def test_normalizing_coordinates(self):
        x = np.arange(5)
        x_norm = np.linspace(-1, 1, len(x))
        fake_data = np.zeros((9, len(x)))
        fitter = SingleTraceFitter(image_data=fake_data, extraargs={'initialize_fit_objects': False})
        fitter._normalize_domain_coordinates()
        assert np.allclose(fitter.x, x)
        assert np.allclose(fitter.x_norm, x_norm)

    def test_generating_legendre_design_matrix(self):
        x = np.arange(5)
        xnorm = np.linspace(-1, 1, len(x))
        design_matrix = np.array([np.ones(len(x)),
                                  xnorm])
        assert np.allclose(design_matrix,
                           SingleTraceFitter._generate_design_matrix(xnorm, poly_fit_order=1))

    @mock.patch('xwavecal.utils.trace_utils.SingleTraceFitter._centers_from_coefficients')
    @mock.patch('xwavecal.utils.trace_utils.optimize.minimize')
    def test_fit_trace(self, minimize, centers):
        coefficients = np.arange(4)
        centers.return_value = None
        minimize.return_value = OptimizeResult(x=coefficients)
        fitter = SingleTraceFitter(extraargs={'initialize_fit_objects': False,
                                              'coefficients': [],
                                              'initial_guess_next_fit': coefficients})

        assert fitter.fit_trace() is None
        assert fitter.coefficients == [coefficients]


class TestTraceMaker:
    def test_properties(self):
        assert TraceMaker(FakeContext()).calibration_type is 'TRACE'

    @pytest.mark.integration
    def test_trace_fit_does_not_crash_on_blank_frame(self):
        order_of_poly_fit = 4
        image = FakeTraceImage(nx=100, ny=100)
        image.set_header_val('RDNOISE', 11)
        noisify_image(image)
        fake_context = FakeContext()
        fake_context.db_address = ''
        trace_fitter = TraceMaker(fake_context)
        trace_fitter.order_of_poly_fit = order_of_poly_fit
        trace_fitter.xmin, trace_fitter.xmax = 50, 60
        trace_fitter.do_stage(image)
        assert True

    @mock.patch('xwavecal.utils.trace_utils.AllTraceFitter.fit_traces')
    def test_trace_maker(self, fit_traces):
        trace_table_name = 'test'
        data = {'id': [1], 'centers': [np.arange(3)]}
        fit_traces.return_value = Trace(data=data)
        expected_trace = Trace(data=data)
        trace_maker = TraceMaker(FakeContext())
        trace_maker.xmin = 5
        trace_maker.xmax = 10
        trace_maker.trace_table_name = trace_table_name
        loaded_trace = trace_maker.do_stage(image=FakeImage())[1]
        assert np.allclose(loaded_trace.get_centers(0), expected_trace.get_centers(0))
        assert np.allclose(loaded_trace.get_id(0), expected_trace.get_id(0))

    @pytest.mark.integration
    def test_accuracy_of_trace_fitting(self):
        """
        test type: Mock Integration Test with metrics for how well trace fitting is doing.
        info: This tests trace making via a blind fit.
        WARNING: Because trace fitting is defined with polynomials which are normalized from -1 to 1, if one squeezes
        the x axis of the image, then the traces bend more drastically. Thus it is recommended you do not change the
        size of the FakeTraceImage.
        """
        read_noise = 11.0
        poly_fit_order = 4

        image = FakeTraceImage()
        image.fiber0_lit, image.fiber1_lit, image.fiber2_lit = False, True, True
        image.set_header_val('RDNOISE', read_noise)

        image, trace_centers, second_order_coefficient_guess = fill_image_with_traces(image,
                                                                                      poly_order_of_traces=poly_fit_order)
        noisify_image(image)
        fake_context = FakeContext()

        trace_maker = TraceMaker(fake_context)
        trace_maker.xmin = image.data.shape[1]//2 - 20
        trace_maker.xmax = image.data.shape[1]//2 + 20
        trace_maker.order_of_poly_fit = poly_fit_order
        trace_maker.second_order_coefficient_guess = second_order_coefficient_guess
        trace = trace_maker.do_stage(image)[1]
        assert trace.data['centers'].shape[0] == trace_centers.shape[0]
        difference = trace.data['centers'] - trace_centers
        logger.debug('median absolute deviation in unit-test trace fitting is {0} pixels'
                     .format(np.median(np.abs(difference - np.median(difference)))))
        logger.debug('standard deviation in unit-test trace fitting is {0} pixels'
                     .format(np.std(difference)))
        logger.debug('worst error (max of true minus found) in unit-test trace fitting is {0} pixels'
                     .format(np.max(np.abs(difference))))
        logger.debug('median error (median of true minus found) in unit-test trace fitting is {0} pixels'
                     .format(np.abs(np.median(difference))))
        assert np.median(np.abs(difference - np.median(difference))) < 1/100
        assert np.abs(np.median(difference)) < 1/100


class TestLoadTrace:
    def test_properties(self):
        assert LoadTrace(FakeContext()).calibration_type is 'TRACE'

    @mock.patch('xwavecal.traces.LoadTrace.get_calibration_filename', return_value=None)
    def test_load_trace_flags_images_without_calibration(self, mock_get_cal):
        fake_context = FakeContext()
        setattr(fake_context, 'db_address', None)
        trace_loader = LoadTrace(fake_context)
        image = trace_loader.do_stage(image=FakeImage())
        assert getattr(image, 'trace') is None

    @mock.patch('xwavecal.traces.LoadTrace.get_calibration_filename', return_value='/path/to/master_trace.fits')
    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('astropy.io.fits.open', return_value=None)
    @mock.patch('xwavecal.utils.trace_utils.Trace.load')
    def test_load_trace(self, mock_load, mock_open, mock_os, mock_get_cal):
        data = {'id': [1], 'centers': [np.arange(3)]}
        expected_trace = Trace(data=data)
        mock_load.return_value = Trace(data=data)
        fake_context = FakeContext()
        setattr(fake_context, 'db_address', None)
        trace_loader = LoadTrace(fake_context)
        image = trace_loader.do_stage(image=FakeImage())
        assert np.allclose(image.trace.get_centers(0), expected_trace.get_centers(0))
        assert np.allclose(image.trace.get_id(0), expected_trace.get_id(0))



import numpy as np


def normalize_orders(image_data, trace, half_window=10, n=100):
    """
    :param image_data: numpy.ndarray
    :param trace: xwavecal.utils.trace_utils.Trace
    :param half_window: int
                        the number of pixels above an below a diffraction order to try and normalize
    :param n: int
              the n brightest pixels of those near the trace will have their median computed, then
              all the pixels near the trace will be divided by that median.
    :return: normalization_factor: numpy.ndarray
             The array that normalizes each diffraction order of image_data to 1.
    """
    maxy = np.max(image_data.shape[0])
    x, y = np.meshgrid(np.arange(image_data.shape[1]), np.arange(image_data.shape[0]))
    normalization_factor = np.ones_like(image_data).astype(float) * np.inf
    for single_order in trace.data['centers']:
        roi = slice(int(max(0, np.min(single_order) - half_window)),
                    int(min(np.max(single_order) + half_window, maxy)))
        near_trace = np.where(np.isclose(y[roi] - single_order, 0, atol=half_window))
        median_of_n_brightest = np.median(np.sort(image_data[roi][near_trace])[::-1][:n])
        normalization_factor[roi][near_trace] = median_of_n_brightest
    return normalization_factor



"""
Module for basic reduction steps like overscan subtraction,
overscan trimming and gain normalization.
"""

import numpy as np
import sep

from xwavecal.stages import Stage
from xwavecal.utils.runtime_utils import import_obj
from xwavecal.utils.basic_utils import median_subtract_channels_y


class OverscanSubtractor(Stage):
    def __init__(self, runtime_context):
        super(OverscanSubtractor, self).__init__(runtime_context)

    def do_stage(self, image):
        parse_region_keyword = import_obj(self.runtime_context.parse_region_keyword)

        self.logger.info('Subtracting by median overscan')
        image.data = np.ascontiguousarray(image.data.astype(float))
        data_section = parse_region_keyword(image.get_header_val('data_section'))
        overscan_section = parse_region_keyword(image.get_header_val('overscan_section'))

        self.logger.info('data section (Y, X) is {0} and overscan section (Y, X) is {1}'.format(data_section, overscan_section))
        image.data[data_section] -= np.median(image.data[overscan_section])
        return image


class GainNormalizer(Stage):
    def __init__(self, runtime_context):
        super(GainNormalizer, self).__init__(runtime_context)

    def do_stage(self, image):
        self.logger.info('Multiplying by gain')
        image.data = image.data.astype(float) * image.get_header_val('gain')
        image.set_header_val('gain', 1.0)
        return image


class Trimmer(Stage):
    def __init__(self, runtime_context):
        super(Trimmer, self).__init__(runtime_context)

    def do_stage(self, image):
        parse_region_keyword = import_obj(self.runtime_context.parse_region_keyword)

        data_section = parse_region_keyword(image.get_header_val('data_section'))
        self.logger.info('Trimming image to (Y, X) {0}'.format(data_section))
        image.data = image.data[data_section]
        image.data = np.ascontiguousarray(image.data)
        return image


class MedianSubtractReadoutsAlongY(Stage):
    def __init__(self, runtime_context):
        super(MedianSubtractReadoutsAlongY, self).__init__(runtime_context)

    def do_stage(self, image):
        self.logger.info('Subtracting the median from each readout channel')
        num = image.get_header_val('num_rd_channels')
        image.data = median_subtract_channels_y(image.data, num)
        return image


class BackgroundSubtract(Stage):
    def __init__(self, runtime_context):
        super(BackgroundSubtract, self).__init__(runtime_context)

    def do_stage(self, image):
        self.logger.info('Background subtracting the 2d frame')
        image.data = image.data.copy(order='C')
        background = sep.Background(image.data).back()
        image.data = image.data - background
        del background
        return image


class BackgroundSubtractSpectrum(Stage):
    def __init__(self, runtime_context=None):
        super(BackgroundSubtractSpectrum, self).__init__(runtime_context=runtime_context)

    def do_stage(self, image):
        self.logger.info('Background subtracting the extracted 1d spectra')
        for key in [self.runtime_context.main_spectrum_name, self.runtime_context.blaze_corrected_spectrum_name]:
            if image.data_tables.get(key) is not None:
                spectrum = image.data_tables[key]
                if len(spectrum) > 0:
                    background = sep.Background(spectrum['flux'].data).back()
                    spectrum['flux'] -= background
                    image.data_tables[key] = spectrum
        return image



import argparse
import glob
import os
import numpy as np
import importlib
from ast import literal_eval

from xwavecal.utils.fits_utils import Translator


def parse_args(args=None):
    parser = argparse.ArgumentParser(description='Reduce an xwavecal spectrograph frame.')
    parser.add_argument("--output-dir", required=True,
                        help="Directory within which to save the processed data files.")
    parser.add_argument("--input-dir", required=False, default=None,
                        help="Directory which contains the raw data files.")
    parser.add_argument('--data-paths', nargs='+', required=False, default=None,
                        help="path(s) to data, usage: '--data_paths path/to/first.fits path/to/second.fits'")
    parser.add_argument("--fpack", required=False, action='store_true',
                        help="fpack output files with the default quantization.")
    parser.add_argument("--config-file", required=True,
                        help="Path to the instrument specific configuration file.")
    parser.add_argument("--frame-type", required=False, default='any',
                        help="Frame type to either fit traces to or wavelength calibrate."
                             "Make sure frame type settings are appropriately set in the config file."
                             "lampflat files are used for tracing, wavecals are wavelength calibration"
                             "frames such as ThAr exposures. Must agree with the frame names in [stages],"
                             "e.g. lampflat, wavecal etc. Ignore to reduce all valid files",
                        type=str.lower)
    args = parser.parse_args(args)
    if args.data_paths is None and args.input_dir is None:
        raise ValueError('both input_dir and data_paths are None. Must specify raw data or a directory of raw data to process.')
    if not os.path.exists(args.config_file):
        raise FileNotFoundError('{0} not found.'.format(args.config_file))
    return args


def get_data_paths(dir_path, files_contain=None):
    all_files = glob.glob(os.path.join(dir_path, '*'))
    return [file for file in all_files if all([item in file for item in files_contain])]


def order_data(data_paths, data_class, primary_ext, header_keys, type_keys):
    translator = Translator(header_keys, type_keys)
    is_not_lampflat = lambda path: 0 if data_class.load(path, primary_ext, translator).get_header_val('type') == 'lampflat' else 1
    data_paths = list(data_paths)
    if len(data_paths) > 0:
        data_paths.sort(key=is_not_lampflat)
    return data_paths


def select_data_of_type(data_paths, data_class, primary_ext, header_keys, type_keys, frame_type='any'):
    is_type = lambda x: x == frame_type
    if frame_type == 'any':
        is_type = lambda x: type(x) is str
    translator = Translator(header_keys, type_keys)
    correct = lambda path: 1 if is_type(data_class.load(path, primary_ext, translator).get_header_val('type')) else 0
    return np.array(data_paths)[np.where([correct(path) for path in data_paths])]


def import_obj(full_class_string):
    """
    dynamically import a class or function from a string
    """

    class_data = full_class_string.split(".")
    module_path = ".".join(class_data[:-1])
    class_str = class_data[-1]

    module = importlib.import_module(module_path)
    return getattr(module, class_str)


def safe_eval(item):
    """
    :param item: str, int or dict
    :return: Any strings have erroneous leading " or ' removed.
    """
    out = literal_eval(item)
    if isinstance(out, str):
        return out.replace("'", '').replace('"', '')
    return out



"""
Module for calculating the inverse variances of the image.
"""

from xwavecal.stages import Stage
import numpy as np


class CalcInverseVariance(Stage):
    def __init__(self, runtime_context=None):
        super(CalcInverseVariance, self).__init__(runtime_context=runtime_context)

    def do_stage(self, image):
        var = estimate_var(image.data, image.get_header_val('read_noise'))
        image.ivar = var ** (-1)
        del var
        return image


def estimate_var(data, rdnoise):
    var = np.copy(data).astype(float)
    var[var < rdnoise ** 2] = 0
    return var + rdnoise ** 2



import numpy as np
import mock
from scipy import optimize

from xwavecal.tests.utils import array_with_peaks
from xwavecal.utils.misc_utils import find_peaks, brute_local_min, find_nearest,\
                                         overlap_region, normalize_by_brightest


def test_find_peaks():
    min_snr = 1
    peak_centers = np.array([15.2, 40.2, 60.2, 80])
    pixel = np.arange(100).astype(float)
    std = 1.5  # standard deviation of the gaussian peaks.
    flux = array_with_peaks(x=pixel, centroids=peak_centers,
                            amplitudes=[1000, 6, min_snr + 0.1, min_snr - 0.1], stds=[std, std, std, std])
    center_uncertainties = std/np.sqrt([np.sum(flux[np.isclose(pixel, cntr, atol=8*std)]) for cntr in peak_centers])
    for min_height in [min_snr, min_snr * np.ones_like(flux)]:
        found_peak_centers, errs = find_peaks(flux, pixel, np.sqrt(flux), height=min_height, distance=10, window=6)[:2]
        assert np.allclose(found_peak_centers, peak_centers[:-1], rtol=.01)
        assert np.allclose(errs, center_uncertainties[:-1], rtol=.1)


@mock.patch('xwavecal.utils.misc_utils.fit_peaks', return_value=np.array([[0.1, 0.01], [13, 2]]))
@mock.patch('scipy.signal.find_peaks', return_value=(np.array([0, 1]), {}))
def test_identify_lines(mock_peaks, mock_refine):
    peak_x, peak_err, peak_ind = find_peaks(None, np.array([0, 20]), None)
    assert np.allclose(peak_x, [0.1, 20])
    assert np.allclose(peak_err, [0.01, 1])


def test_brute_force_min():
    def dimple(x, dummy):
        return -x**2 - 50 * np.exp(-x ** 2/(2*0.3**2)) + 500
    root = optimize.minimize(dimple, x0=1.01, method=brute_local_min, args=None,
                             options={'step': 1/10, 'rrange': (-15, 15), 'filtw': 31, 'finish': None}).x
    assert np.isclose(root[0], 0, atol=0.1)
    root = optimize.minimize(dimple, x0=1.01, method=brute_local_min, args=None,
                             options={'step': 1/10, 'rrange': (-15, 15), 'filtw': 31, 'finish': 'Nelder-Mead'}).x
    assert np.isclose(root[0], 0)


class TestFindNearest:
    # the find_nearest function is at the heart of the wavelength solution and warrants exhaustive testing.
    @staticmethod
    def find_nearest_slow(array_a, array_b):
        nearest_value_idxs = [(np.abs(array_b - value)).argmin() for value in array_a]
        return array_b[nearest_value_idxs]

    def test_strenuous(self):
        b = np.sort(np.random.random(size=100))
        a = np.random.random(size=100) * 2 - 1
        assert np.allclose(self.find_nearest_slow(a, b), find_nearest(a, b))

    def test_simple(self):
        b = np.array([1, 2, 3])
        a = np.array([1.1, 1.9, 3.49])
        assert np.allclose(self.find_nearest_slow(a, b), b)
        assert np.allclose(find_nearest(a, b), b)


def test_overlap_region():
    assert np.allclose(overlap_region([1, 2], [1.5, 3]), (1.5, 2))
    assert overlap_region([1, 2], [2.001, 3]) is None


def test_normalize_by_brightest():
    spec = normalize_by_brightest(np.array([[1, 2, 3, 2], [1, 1, 1, 1]], dtype=float), n=3)
    assert np.allclose(spec, [[.5, 1, 3/2, 1], [1, 1, 1, 1]])



from setuptools import setup, find_packages

setup(name='xwavecal',
      author='G. Mirek Brandt, Curtis McCully, Timothy Brandt',
      author_email='gmbrandt@ucsb.edu',
      version='0.1.11',
      python_requires='>=3.6',
      url='https://github.com/gmbrandt/xwavecal',
      packages=find_packages(),
      package_dir={'xwavecal': 'xwavecal'},
      package_data={'xwavecal': ['data/.*']},
      setup_requires=['pytest-runner'],
      install_requires=[
                        'astropy<=4.0',
                        'scipy<=1.3.2',
                        'sep<=1.10.0',
                        'numpy>=1.16',
      ],
      entry_points={'console_scripts': ['xwavecal_reduce_dir=xwavecal.main:run',
                                        'xwavecal_reduce=xwavecal.main:reduce_data']})



import numpy as np

from xwavecal.utils.misc_utils import find_peaks, find_nearest
from astropy.stats import sigma_clip


class Model(dict):
    def add_polynomial_term(self, x_degree, i_degree):
        """
        :param self: dict
        :param x_degree: int, the degree of the x term to add to the polynomial model
        :param i_degree: int, the degree of the order term to add to the polynomial model
        :return: model such that model[x_degree] is a list with i_degree in the list.
        """
        if x_degree not in self:
            self[x_degree] = []
        if i_degree not in self[x_degree]:
            self[x_degree].append(i_degree)

    def is_missing_polynomial_term(self, x_degree, i_degree):
        if x_degree not in self or i_degree not in self[x_degree]:
            return True
        return False


def normalize_coordinates(coordinates, max_value, min_value):
    """
    :param coordinates: list or ndarray
    :param max_value: max_value coordinates can achieve. e.g. if normalizing pixels, we would have 4096 as max_value
    :return: coordinates normalized to run from -1 to 1.
    """
    coordinates = np.array(coordinates)
    if min_value > max_value:
        raise ValueError('min_value > max_value')
    coordinates = 2. * (coordinates - min_value)/(max_value - min_value) - 1.
    return coordinates


def estimate_global_scale(detector_range, n, m0):
    """
    :param detector_range: max wavelength minus min wavelength of the detector in angstroms, e.g. 5000
    :param n: the number of distinct diffraction orders on the detector, e.g. 67
    :param m0: the principle order number, e.g. 52
    :return: The global scale constant for the wavelength solution, in units of Angstrom.
    """
    return detector_range * (1/m0 - 1/(m0 + n))**(-1)


def identify_lines(spectrum, stderr, min_snr=5, min_ptp=5, order_key='ref_id'):
    """
    :param spectrum: Table or dictionary with 'flux', 'pixel' and some order column
    :param stderr: float, array with shape (len(spectrum['flux']), 1),
                   or array with shape spectrum['flux'].shape : specifying the standard error globally,
                   per order or per pixel respectively.
    :param min_snr: min signal to noise ratio that a peak must possess.
    :param min_ptp: min peak-to-peak spacing in pixels
    :param order_key: the dict key such that spectrum[order_key] gives a column of order id's.
    :return: dictionary
             with keys 'pixel', 'order' and 'flux' giving 1d arrays of the peak (emission lines)
             locations pixel, order and their peak flux.
    """
    lines = {'order': [], 'pixel': [], 'pixel_err': [], 'flux': []}
    std_err = np.ones((len(spectrum), 1)) * stderr
    for row in range(len(spectrum)):
        peak_coordinates, peak_err, peak_indices = find_peaks(spectrum['flux'][row], spectrum['pixel'][row],
                                                              yerr=std_err[row], height=std_err[row] * min_snr,
                                                              distance=min_ptp,
                                                              prominence=0.5 * np.abs(spectrum['flux'][row]),
                                                              peak_width=2, window=6)
        if len(peak_indices) > 0:
            lines['flux'].extend(list(spectrum['flux'][row][peak_indices]))
            lines['pixel'].extend(list(peak_coordinates))
            lines['pixel_err'].extend(list(peak_err))
            lines['order'].extend([spectrum[order_key][row]]*len(peak_coordinates))
    # casting as arrays for convenient use.
    lines['pixel'] = np.array(lines['pixel'], dtype=np.float32)
    lines['pixel_err'] = np.array(lines['pixel_err'], dtype=np.float32)
    lines['order'] = np.array(lines['order'], dtype=np.int)
    lines['flux'] = np.array(lines['flux'], dtype=np.float32)
    return lines


def calc_residuals(wavelengths, reference_wavelengths):
    closest_ref_wavelengths = find_nearest(wavelengths, array_b=reference_wavelengths)
    return wavelengths - closest_ref_wavelengths


def restrict(lines, low_order=-np.inf, high_order=np.inf):
    if low_order == -np.inf and high_order == np.inf:
        return lines
    order_mask = np.logical_and(lines['order'] <= high_order, lines['order'] >= low_order)
    restricted_lines = {key: coordinates[order_mask] for key, coordinates in lines.items()}
    return restricted_lines


def _sigma_clip(residuals, lines, sigma=3, stdfunc=np.std):
    if len(residuals) == 0:
        return lines
    clip_mask = ~ sigma_clip(residuals, sigma=sigma, stdfunc=stdfunc, masked=True).mask
    lines = {key: coordinates[clip_mask] for key, coordinates in lines.items()}
    return lines


def pixel_order_as_array(spectrum):
    # TODO check if you NEED to call .data here, I don't think you do.
    pixels = spectrum['pixel'].data
    orders = np.repeat(spectrum['ref_id'].data.reshape(-1, 1), pixels.shape[1], axis=1)
    return pixels, orders



"""
trace_utils.py: Routines for finding xwavecal orders across a CCD.
Authors
    G. Mirek Brandt (gmbrandt@ucsb.edu)

    Tim Brandt (tbrandt@physics.ucsb.edu)

    Curtis McCully (cmccully@lcogt.net)
"""


import numpy as np
from scipy import ndimage, optimize, signal
from astropy.table import Table, Column
from xwavecal.images import DataProduct

import logging
logger = logging.getLogger(__name__)


class Trace(DataProduct):
    """
    :param data = {'id': ndarray, 'centers': ndarray}. 'centers' gives a 2d array, where
    the jth row are the y centers across the detector for the trace with identification trace_centers['id'][j]
    """
    def __init__(self, data=None, table_name=None, num_centers_per_trace=0, filepath=None,
                 header=None, translator=None):
        super(Trace, self).__init__(data=data, data_name=table_name, filepath=filepath,
                                    header=header, translator=translator)
        if data is None and num_centers_per_trace <= 0:
            raise ValueError('Trace object instantiated but no trace data given and num_centers_per_trace is not > 0')
        if data is None:
            data = Table([Column(name='id'), Column(name='centers', shape=(num_centers_per_trace,))])
            data['id'].description = 'Identification tag for trace'
            data['centers'].description = 'Vertical position of the center of the' \
                                          ' trace as a function of horizontal pixel'
            data['centers'].unit = 'pixel'
        self.data = Table(data)

    def get_centers(self, row):
        return self.data['centers'][row]

    def get_id(self, row):
        return self.data['id'][row]

    def add_centers(self, trace_centers, trace_id):
        self.data.add_row([trace_id, trace_centers])

    def del_centers(self, rows):
        self.data.remove_rows(rows)

    def num_traces_found(self):
        return len(self.data['id'])

    def sort(self):
        center = int(self.data['centers'].shape[1] / 2)
        self.data['centers'] = np.array(self.data['centers'])
        self.data['centers'] = self.data['centers'][self.data['centers'][:, center].argsort()]
        self.data['id'] = np.arange(self.data['centers'].shape[0])

    def remove_duplicates(self, thresh=3):
        c = int(self.data['centers'].shape[1] / 2)
        duplicates = np.arange(len(self.data['centers']) - 1)[np.isclose(self.data['centers'][1:, c],
                                                                         self.data['centers'][:-1, c], atol=thresh)]
        self.del_centers(duplicates)
        self.sort()


class AllTraceFitter(object):
    def __init__(self, xmin=2000, xmax=2100, min_peak_to_peak_spacing=10, min_snr=6):
        self.xmin = xmin
        self.xmax = xmax
        self.min_peak_to_peak_spacing = min_peak_to_peak_spacing
        self.min_snr = min_snr

    def fit_traces(self, trace, image_data, poly_fit_order, second_order_coefficient_guess, image_noise_estimate=10):
        """
        :param trace: Trace object with get_centers, num_traces_found, add_centers, del_centers, and sort method
        :param image_data: The image data array, e.g. image.data or hdu[1].data for a fits HDUlist
        :param poly_fit_order: degree of the polynomial which will fit the xwavecal orders across the detector
        :param second_order_coefficient_guess: guess for the 2nd order coefficient of the aforementioned polynomial
        :param image_noise_estimate: estimate of the images read noise.
        :return: trace object with the y coordinates of the trace centers as a function of x pixel.
        """
        trace_fitter = SingleTraceFitter(image_data=image_data,
                                         second_order_coefficient_guess=second_order_coefficient_guess,
                                         poly_fit_order=poly_fit_order)
        peak_xy_coordinates = self._identify_traces(image_data, image_noise_estimate)
        trace = self._step_through_detector(trace, trace_fitter, peak_xy_coordinates)
        trace.sort()
        trace.remove_duplicates(thresh=self.min_peak_to_peak_spacing/2)
        return trace

    @staticmethod
    def _step_through_detector(trace, trace_fitter, peak_xy_coordinates):
        for trace_id, peak_xy_coordinate in enumerate(peak_xy_coordinates):
            trace_fitter.update_initial_guess_to_run_through_pt(peak_xy_coordinate)
            trace_centers = trace_fitter.fit_trace()
            trace.add_centers(trace_centers, trace_id)
            trace_fitter.use_fit_as_initial_guess(-1)
        return trace

    def _identify_traces(self, bkg_subtracted_image_data, image_noise_estimate):
        """
        :param bkg_subtracted_image_data: Image with the background subtracted
        :param image_noise_estimate: Read noise on the image.
        :return:
        """
        flux_down_detector = self._filtered_flux_down_detector(bkg_subtracted_image_data, image_noise_estimate)
        peak_y_coordinates = signal.find_peaks(flux_down_detector, distance=self.min_peak_to_peak_spacing,
                                               height=self.min_snr)[0]
        peak_x_coordinates = np.ones_like(peak_y_coordinates) * np.mean((self.xmin, self.xmax), dtype=int)
        peak_xy_coordinates = list(zip(peak_x_coordinates, peak_y_coordinates))
        return peak_xy_coordinates

    def _filtered_flux_down_detector(self, image_data, image_noise_estimate):
        wedge_of_image = image_data[:, self.xmin:self.xmax]
        noise_normalized_wedge = wedge_of_image / np.sqrt(image_noise_estimate**2 + np.abs(wedge_of_image))
        median_slice = np.median(noise_normalized_wedge, axis=1)
        return median_slice


class SingleTraceFitter(object):
    def __init__(self, image_data=None, poly_fit_order=2, second_order_coefficient_guess=None,
                 extraargs=None):
        if extraargs is None:
            extraargs = {}
        if extraargs.get('coefficients') is None:
            extraargs['coefficients'] = []
        self.second_order_coefficient_guess = second_order_coefficient_guess
        self.image_data = image_data
        self.poly_fit_order = poly_fit_order
        self.filtered_image_data = extraargs.get('filtered_image_data')
        self.initial_guess_next_fit = extraargs.get('initial_guess_next_fit')
        self.coefficients = extraargs.get('coefficients')
        self.x = extraargs.get('x')
        self.x_norm = extraargs.get('xnorm')
        self.design_matrix = extraargs.get('design_matrix')

        if extraargs.get('initialize_fit_objects', True):
            if self.second_order_coefficient_guess is None:
                logger.error('The second order guess have been specified '
                             ', aborting trace fitting.')
                raise ValueError('Starting y position up the detector nor the second order guess have been specified')
            self._normalize_domain_coordinates()
            self.filtered_image_data = self._prefilter_image_data(self.image_data)
            self.design_matrix = self._generate_design_matrix(self.x_norm, self.poly_fit_order)
            self.initial_guess_next_fit = np.zeros(self.poly_fit_order + 1).astype(np.float64)
            self.initial_guess_next_fit[2] = self.second_order_coefficient_guess

    def fit_trace(self):
        refined_coefficients = optimize.minimize(SingleTraceFitter._trace_merit_function, self.initial_guess_next_fit,
                                                 args=(self,), method='Powell').x
        self.coefficients.append(refined_coefficients)
        return self._centers_from_coefficients(refined_coefficients)

    def use_fit_as_initial_guess(self, fit_id):
        self.initial_guess_next_fit = np.copy(self.coefficients[fit_id])

    @staticmethod
    def _generate_design_matrix(normalized_domain, poly_fit_order):
        design_matrix = np.ones((poly_fit_order + 1, normalized_domain.shape[0]))
        for i in range(poly_fit_order + 1):
            design_matrix[i] = legendre(order=i, values=normalized_domain)
        return design_matrix

    def _centers_from_coefficients(self, coefficients):
        """
        :param coefficients: a 1d array for the coefficients of a single trace in order of: [0th, 1st,.. mth degree coefficient]
        or a 2d array of shape (N, m+1) for N traces.
        :return: If coefficients is 1d of size m+1, this will return an array of shape image_data.shape[1] with the trace
        centers at each of those points. If coefficients is for many traces (shape (N, m+1)) then this will return a 2d
        array of shape (N, image_data.shape[1]) e.g. the trace centers for each of the traces.
        """
        trace_centers = np.dot(coefficients, self.design_matrix)
        return trace_centers

    def update_initial_guess_to_run_through_pt(self, xy):
        """
        :param xy: (x,y) point on the detector you want the initial_guess for the next fit to pass through. (0,0)
                    is the smallest coordinate as usual.
                    This method does not change the shape of the trace generated by initial_guess_next_fit, it only
                    shifts the trace so that at the pixel x, the trace center has a y value of y.
        """
        # NOTE: This method calls all the trace centers, which means it evalutes 4096 centers, when it only needs one.
        current_centers = self._centers_from_coefficients(self.initial_guess_next_fit)
        self.initial_guess_next_fit[0] += xy[1] - current_centers[xy[0]]

    def _flux_across_trace(self, trace_centers):
        total_flux = np.sum(ndimage.map_coordinates(self.filtered_image_data, [trace_centers, self.x], prefilter=False))
        return total_flux

    def _normalize_domain_coordinates(self):
        self.x = np.arange(self.image_data.shape[1])
        # note this normalization scheme only works if np.min(x) = 0, which it is in this case.
        self.x_norm = self.x * 2. / self.x[-1] - 1  # x normalized to run from -1 to 1

    @staticmethod
    def _trace_merit_function(single_trace_coefficients, cls):
        trace_centers = cls._centers_from_coefficients(single_trace_coefficients)
        flux = cls._flux_across_trace(trace_centers)
        return (-1)*flux

    @staticmethod
    def _prefilter_image_data(image_data):
        return ndimage.spline_filter(image_data)


def legendre(order, values):
    #TODO replace with scipy.legendre(order)(values) if it works the same.
    if order == 0:
        return np.ones_like(values)
    else:
        return np.polynomial.legendre.legval(values, [0 for j in range(order)] + [1])



from scipy import signal
import numpy as np


def correlate2d(array, template, max_lag=100):
    """
    wrapper for signal.correlate2d in which you can specify a max offset for the correlation.
    works by padding the input array with zeros and enforcing 'valid' mode in signal.correlate 2d so
    that the template cannot slide off the array.
    Both dimensions of array must be larger than template.
    -----
    :param array: 2d ndarray to correlate with the template.
    :param template: 2d ndarray.
    :param max_lag: The maximum correlation offset in either dimension.
    :return 2d cross correlation signal as a function of offset.
    """
    top_pad = int(template.shape[0] / 2)
    temp_array = np.pad(array, pad_width=((top_pad, top_pad), (max_lag, max_lag)),
                        mode='constant', constant_values=0)
    return signal.correlate2d(temp_array, template, mode='valid')



import pytest
import mock
import numpy as np
from scipy.optimize import OptimizeResult
import copy
from astropy.table import Table

import xwavecal.utils.wavelength_utils as wcsu
from xwavecal.utils.trace_utils import legendre as legendre_val
from xwavecal.utils.overlap_utils import blank_overlap_table
from xwavecal.wavelength import WavelengthSolution, FindGlobalScale, SolutionRefineInitial, SolutionRefineFinal
from xwavecal.wavelength import refine_wcs, FitOverlaps, WavelengthStage, SolveFromOverlaps, IdentifyArcEmissionLines
from xwavecal.wavelength import ApplyToSpectrum, TabulateArcEmissionLines, BlazeCorrectArcEmissionLines, Initialize
from xwavecal.wavelength import IdentifyPrincipleOrderNumber, SolutionRefineOnce, find_feature_wavelengths
from xwavecal.tests.utils import SpectrumUtils, FakeImage, FakeContext


class TestWavelengthSolution:
    def test_format_overlaps(self):
        overlaps = Table({'pixel': [np.arange(3, 6), np.arange(3, 6)], 'id': [1, 2]})
        wcs = WavelengthSolution(min_order=1, max_order=2, min_pixel=3, max_pixel=5)
        coordinates = wcs._format_overlaps(overlaps, pixel_key='pixel', order_key='id')
        assert np.allclose(coordinates['normed_order'], [-1, -1, -1, 1, 1, 1])
        assert np.allclose(coordinates['normed_pixel'], [-1, 0, 1, -1, 0, 1])
        assert np.allclose(coordinates['order'], [1, 1, 1, 2, 2, 2])

    def test_format_overlaps_rejects_nan(self):
        overlaps = Table({'pixel': [np.array([3, 4, np.nan]), np.array([np.nan, 4, 5])], 'id': [1, 2]})
        wcs = WavelengthSolution(min_order=1, max_order=2, min_pixel=3, max_pixel=5)
        coordinates = wcs._format_overlaps(overlaps, pixel_key='pixel', order_key='id')
        assert np.allclose(coordinates['normed_order'], [-1, -1, 1, 1])
        assert np.allclose(coordinates['normed_pixel'], [-1, 0, 0, 1])
        assert np.allclose(coordinates['order'], [1, 1, 2, 2])

    @pytest.mark.integration
    @mock.patch('xwavecal.wavelength.WavelengthSolution._format_overlaps')
    def test_solve_from_overlaps(self, mock_format):
        m0 = 52
        m_coordinates = {'normed_pixel': np.array([1]), 'normed_order': np.array([1]), 'order': np.array([21])}
        coordinates = {'normed_pixel': np.array([-1]), 'normed_order': np.array([-1]), 'order': np.array([20])}

        def fake_format(overlaps, pixel_key=None, order_key=None):
            return m_coordinates if pixel_key == 'matched_pixel' else coordinates
        mock_format.side_effect = fake_format
        wcs = WavelengthSolution(m0=m0, model={1: [0]})  # (1 + b * x)/(m0 + order_num) model

        wcs.solve_from_overlaps(None)
        coefficients = wcs.model_coefficients
        expected_coefficient = 1/(legendre_val(1, 1)/(m0 + 21) - legendre_val(1, -1)/(m0 + 20)) *\
                                                                 (1/(m0 + 20) - 1/(m0 + 21))
        assert np.isclose(coefficients[0], expected_coefficient)

    def test_invalid_model_raises_error(self):
        with pytest.raises(ValueError):
            WavelengthSolution(model={0: [0]}).solve_from_overlaps(None)

    def test_construct_wavelength_map_matrices_grating_eq(self):
        m0 = 52
        coordinates = {'normed_pixel': np.array([-1]), 'normed_order': np.array([-1]), 'order': np.array([20])}
        wcs = WavelengthSolution(m0=m0, model={1: [0]})  # (1 + b * x)/(m0 + order_num) model
        A, c = wcs._construct_wavelength_map_matrices(coordinates['normed_pixel'], coordinates['normed_order'],
                                                      coordinates['order'])
        assert np.allclose(c, [[-1/(m0 + 20)]])
        assert np.allclose(A, [[1/(m0 + 20) * coordinates['normed_pixel'][0]]])

    def test_construct_wavelength_map_matrices_no_grating(self):
        m0 = 52
        coordinates = {'normed_pixel': np.array([-1]), 'normed_order': np.array([-1]), 'order': np.array([20])}
        wcs = WavelengthSolution(m0=m0, model={1: [0]}, grating_eq=False)  # (1 + b * x) model
        A, c = wcs._construct_wavelength_map_matrices(coordinates['normed_pixel'], coordinates['normed_order'],
                                                      coordinates['order'])
        assert np.allclose(c, [[-1]])
        assert np.allclose(A, [[1 * coordinates['normed_pixel'][0]]])

    def test_construct_wavelength_map_matrices_no_force_unity(self):
        m0 = 52
        coordinates = {'normed_pixel': np.array([-1]), 'normed_order': np.array([-1]), 'order': np.array([20])}
        wcs = WavelengthSolution(m0=m0, model={0: [0]})  # (1 + b * x)/(m0 + order_num) model
        A, c = wcs._construct_wavelength_map_matrices(coordinates['normed_pixel'], coordinates['normed_order'],
                                                      coordinates['order'])
        assert np.allclose(c, [[0]])

    def test_solve(self):
        coordinates = {'normed_pixel': np.array([-1, 1]), 'normed_order': np.array([1, 1]),
                       'order': np.array([20, 20])}
        wcs = WavelengthSolution(m0=52, model={0: [0], 1: [1]})  # (a + b*x*i)/(m0+i) model
        a, b = 1.31, 5.42
        wavelengths = 1/(52 + 20) * np.array([a + b * (-1)*1, a + b * 1*1])
        coeffs = wcs.solve(coordinates, wavelengths)
        assert np.allclose(coeffs, [a, b])

    def test_weighted_solve(self):
        coordinates = {'normed_pixel': np.array([-1, 1, .5]), 'normed_order': np.array([1, 1, .5]),
                       'order': np.array([20, 20, 22])}
        wcs = WavelengthSolution(m0=52, model={0: [0], 1: [1]})  # (a + b*x*i)/(m0+i) model
        a, b = 1.31, 5.42
        wavelengths = 1/(52 + 20) * np.array([a + b * (-1)*1, a + b * 1*1, 10])
        coeffs = wcs.solve(coordinates, wavelengths, weights=np.array([1, 1, 0]))
        assert np.allclose(coeffs, [a, b])

    def test_update_model(self):
        wcs = WavelengthSolution(m0=52, model={0: [0], 1: [1, 2]}, model_coefficients=np.arange(1, 4),
                                 min_order=0, max_order=5, min_pixel=0, max_pixel=400)
        new_model = {0: [0, 1], 1: [1, 2, 3]}
        wcs.update_model(new_model=new_model)
        assert new_model == wcs.model
        assert np.allclose(np.array([1, 0, 2, 3, 0]), wcs.model_coefficients)

    def test_update_model_complex(self):
        model = {0: [0, 1, 2, 3, 4, 5], 1: [0, 1, 2, 3, 4, 5], 2: [0, 1, 2, 3, 4, 5], 3: [0, 1, 2, 3, 4]}
        wcs = WavelengthSolution(m0=52, model=model, model_coefficients=np.ones(23),
                                 min_order=0, max_order=67, min_pixel=0, max_pixel=4095)
        new_model = wcsu.Model(copy.deepcopy(model))
        new_model.add_polynomial_term(3, 5)
        wcs.update_model(new_model=new_model)
        assert new_model == wcs.model
        assert np.allclose([1]*23 + [0], wcs.model_coefficients)

    def test_apply_scale(self):
        wcs = WavelengthSolution(model_coefficients=np.arange(5))
        wcs.apply_scale(2)
        assert (wcs.model_coefficients, np.arange(5) * 2)

    def test_wavelength_is_nan_for_no_coefficients(self):
        wcs = Utils.simple_wcs()
        wcs.model_coefficients = None
        assert np.all(np.isnan(wcs(np.arange(5), np.arange(5))))


class TestModel:
    def test_add_term(self):
        model = wcsu.Model({0: [0]})
        model.add_polynomial_term(x_degree=3, i_degree=2)
        assert model == wcsu.Model({0: [0], 3: [2]})

    def test_check_term_exists(self):
        assert not wcsu.Model({0: [1]}).is_missing_polynomial_term(0, 1)
        assert wcsu.Model({0: [1]}).is_missing_polynomial_term(0, 0)


class TestStatistics:
    @mock.patch('xwavecal.utils.wavelength_utils.find_nearest', return_value=np.array([1, 2.]))
    def test_calc_residuals(self, fake_match):
        assert np.allclose(wcsu.calc_residuals(np.array([1., 3]), None), [0, 1.0])
        assert np.allclose(wcsu.calc_residuals(np.array([1., 2]), None), [0, 0])


class TestWavelengthStage:
    def test_do_stage(self):
        image = FakeImage()
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 1, 0
        image.wavelength_solution = {0: None, 1: 'placeholder', 2: None}
        image = WavelengthStage(FakeContext()).do_stage(image)
        assert image.wavelength_solution[1] == 'placeholder'

    @mock.patch('xwavecal.wavelength.WavelengthStage.do_stage_fiber', return_value=None)
    def test_do_stage_skips_if_no_valid_fibers(self, fake_fiber_stage):
        image = FakeImage()
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 0, 1
        image.wavelength_solution = {0: None, 1: 'placeholder', 2: None}
        assert WavelengthStage(FakeContext()).do_stage(image) is not None

    def test_valid_fibers(self):
        image = FakeImage()
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 1, 0
        image.wavelength_solution = {0: None, 1: 'placeholder', 2: None}
        valid_fibers = WavelengthStage._valid_fibers(image)
        assert np.allclose(valid_fibers, [1])
        image.wavelength_solution = {0: None, 1: None, 2: 'placeholder'}
        assert len(WavelengthStage._valid_fibers(image)) == 0


class TestInitialize:
    CONTEXT = FakeContext()

    def test_on_no_valid_fibers(self):
        image = FakeImage()
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 1, 1
        image.wavelength_solution = {1: 'not none', 2: 'not none'}
        Initialize(None).on_no_valid_fibers(image)
        assert image.wavelength_solution == {1: None, 2: None}

    def test_all_fibers_invalid_without_spectrum(self):
        image = FakeImage()
        image.data_tables = {}
        assert len(Initialize(self.CONTEXT)._valid_fibers(image)) == 0

    def test_all_fibers_invalid_without_ref_id_and_fiber_cols(self):
        image = FakeImage()
        image.data_tables[self.CONTEXT.main_spectrum_name] = Table({'col1': [1]})
        assert len(Initialize(self.CONTEXT)._valid_fibers(image)) == 0


class TestFitOverlaps:
    @mock.patch('xwavecal.wavelength.fit_overlaps')
    def test_do_stage(self, mock_overlaps):
        num = 2
        image = FakeImage()
        context = FakeContext()
        spectrum = Table({'ref_id': [0, 1], 'fiber': [0, 1], 'flux': [[0, 1, 2], [0, 1, 2]],
                          'pixel': [[-1, 0, 1], [-1, 0, 1]]})
        lines = Table({'order': [0, 1], 'fiber': [0, 1], 'flux': [[0, 1, 2], [0, 1, 2]],
                       'pixel': [[-1, 0, 1], [-1, 0, 1]]})
        mock_overlaps.return_value = Table({'fiber': [1, 0], 'ref_id': [0, 1], 'matched_ref_id': [1, 2],
                                            'good': [True, True], 'peaks': [10, 10]})
        image.data_tables = {context.main_spectrum_name: spectrum}
        wcs = WavelengthSolution(min_order=0, max_order=num, min_pixel=0, max_pixel=500, model={1: [1], 2: [1]})
        image.wavelength_solution = {0: wcs, 1: wcs}
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 1, 1, 0
        image = FitOverlaps(FakeContext()).do_stage(image)
        assert isinstance(image.data_tables[context.overlap_table_name], Table)

    @mock.patch('xwavecal.wavelength.fit_overlaps')
    def test_do_stage_single_fiber_stacks_overlap_tables(self, fake_fit):
        image = FakeImage()
        context = FakeContext()
        image.data_tables = {context.overlap_table_name: Table({'ref_id': [0], 'fiber': [0], 'matched_ref_id': [1],
                                                                      'good': [False], 'peaks': [10]})}
        image.data_tables[context.main_spectrum_name] = Table({'fiber': [0, 1]})
        wcs = WavelengthSolution(min_order=0, max_order=10, min_pixel=0, max_pixel=500, model={1: [1], 2: [1]})
        image.wavelength_solution = {0: wcs, 1: wcs}
        fake_fit.return_value = Table({'ref_id': [0], 'fiber': [1], 'matched_ref_id': [1], 'good': [False], 'peaks': [2]})
        image = FitOverlaps(FakeContext()).do_stage_fiber(image, fiber=1)
        assert len(image.data_tables[context.overlap_table_name]['fiber']) == 2

    @mock.patch('xwavecal.wavelength.fit_overlaps', return_value=blank_overlap_table(3))
    def test_do_stage_sets_wcs_to_none(self, fake_overlaps):
        image = FakeImage()
        context = FakeContext()
        image.data_tables = {context.main_spectrum_name: Table({'ref_id': [0, 1],
                                                                     'fiber': [0, 1],
                                                                     'flux': [[0, 1, 2], [0, 1, 2]],
                                                                     'pixel': [[-1, 0, 1], [-1, 0, 1]]}),
                             context.overlap_table_name: blank_overlap_table(3)}
        image.wavelength_solution = {0: WavelengthSolution()}
        image = FitOverlaps(FakeContext()).do_stage_fiber(image, fiber=0)
        assert image.wavelength_solution[0] is None


class TestSolveFromOverlaps:
    @mock.patch('xwavecal.wavelength.WavelengthSolution.solve_from_overlaps')
    def test_do_stage_does_not_crash(self, fake_solve):
        image = FakeImage()
        context = FakeContext()
        context.min_num_overlaps = 0
        image.data_tables = {context.overlap_table_name: Table({'ref_id': [0, 1],
                                                                     'matched_ref_id': [1, 2],
                                                                     'good': [True, True],
                                                                     'fiber': [0, 0]})}
        image.wavelength_solution = {0: WavelengthSolution()}
        image = SolveFromOverlaps(FakeContext()).do_stage_fiber(image, fiber=0)
        assert np.allclose(image.wavelength_solution[0].overlap_range, (0, 2))

    @mock.patch('xwavecal.wavelength.WavelengthSolution.solve_from_overlaps')
    def test_do_stage_does_not_overwrite_overlaps(self, fake_solve):
        image = FakeImage()
        context = FakeContext()
        context.min_num_overlaps = 0
        image.data_tables = {context.overlap_table_name: Table({'fiber': [1, 0],
                                                                     'ref_id': [0, 1],
                                                                     'matched_ref_id': [1, 2],
                                                                     'good': [True, True]})}
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 1, 1, 0
        image.wavelength_solution = {0: Utils.simple_wcs(), 1: Utils.simple_wcs()}
        image = SolveFromOverlaps(FakeContext()).do_stage(image)
        assert np.allclose(image.data_tables[context.overlap_table_name]['fiber'], [1, 0])


class TestIdentifyArcEmissionLines:
    @pytest.mark.integration
    @mock.patch('xwavecal.wavelength.identify_lines')
    def test_do_stage_fiber(self, fake_measured_lines):
        fake_measured_lines.return_value = {'pixel': np.array([0, 20]), 'order': np.array([0, 5]), 'pixel_err': np.array([1, 2])}
        image = FakeImage()
        image.wavelength_solution = {1: WavelengthSolution(min_pixel=0, max_pixel=20, min_order=-5,
                                                           max_order=5)}
        image.data_tables = {FakeContext().main_spectrum_name: Table({'fiber': np.array([0, 1]),
                                                                     'stderr': [0, 0]})}
        stage = IdentifyArcEmissionLines(FakeContext())
        image = stage.do_stage_fiber(image, fiber=1)
        measured_lines = image.wavelength_solution[1].measured_lines
        assert np.allclose([measured_lines['pixel'], measured_lines['order']], [[0, 20], [0, 5]])
        assert np.allclose([measured_lines['normed_pixel'], measured_lines['normed_order']],
                           [[-1, 1], [0, 1]])


class TestFindGlobalScale:
    @mock.patch('xwavecal.wavelength.WavelengthSolution.update_model')
    @mock.patch('xwavecal.wavelength.FindGlobalScale._find_scale', return_value=10)
    @mock.patch('xwavecal.wavelength.estimate_global_scale', return_value=1)
    def test_do_stage_fiber(self, fake_guess, fake_result, fake_change):
        image = FakeImage()
        image.wavelength_solution = {1: WavelengthSolution(model_coefficients=1)}
        image = FindGlobalScale(FakeContext()).do_stage_fiber(image, fiber=1)
        assert np.isclose(image.wavelength_solution[1].model_coefficients, 10)

    def test_estimate_global_scale(self):
        assert np.isclose(wcsu.estimate_global_scale(5000, n=67, m0=52), 461791, atol=2)
        assert np.isclose(wcsu.estimate_global_scale(6000, n=67, m0=50), 523880, atol=2)

    @mock.patch('xwavecal.wavelength.calc_residuals', return_value=np.arange(10))
    def test_chi_squared_many_scales(self, fake_chi_squared):
        chi_squared = FindGlobalScale._chi_squared(scales=np.array([1, 2]), unscaled_wavelengths=np.ones(5),
                                                   reference_wavelengths=None)
        assert chi_squared.shape == (2,)
        assert np.allclose(chi_squared[0], np.sum(np.arange(5) ** 2))
        assert np.allclose(chi_squared[1], np.sum(np.arange(5, 10) ** 2))

    def test_chi_squared_safe(self):
        ref = 5 * np.random.random(30)
        chi_squared_s = FindGlobalScale._chi_squared_safe(scales=np.arange(5), unscaled_wavelengths=np.ones(5),
                                                          reference_wavelengths=ref, mem_limit=32)
        chi_squared = FindGlobalScale._chi_squared(scales=np.arange(5), unscaled_wavelengths=np.ones(5),
                                                   reference_wavelengths=ref)
        assert np.allclose(chi_squared, chi_squared_s)

    @mock.patch('xwavecal.wavelength.calc_residuals', return_value=np.arange(5))
    def test_chi_squared_single_scale(self, fake_chi_squared):
        chi_squared = FindGlobalScale._chi_squared(scales=1, unscaled_wavelengths=np.ones(5),
                                                   reference_wavelengths=None)
        assert chi_squared.shape == (1,)
        assert np.allclose(chi_squared[0], np.sum(np.arange(5) ** 2))

    @mock.patch('xwavecal.wavelength.WavelengthSolution.wavelength_normed_input', return_value=1)
    @mock.patch('xwavecal.wavelength.restrict')
    @mock.patch('xwavecal.wavelength.optimize.minimize')
    def test_find_scale(self, fake_minimize, fake_restrict, fake_wavelength):
        wcs = WavelengthSolution(reference_lines=np.arange(10)[::-1])

        def check_minimize_args(fun=None, args=(), x0=None, method=None, options=None, **kwargs):
            args_passed = (x0 == 'bla' and np.allclose(args[1], np.arange(10)) and args[0] == 1)
            return OptimizeResult(x=[args_passed])
        fake_minimize.side_effect = check_minimize_args
        assert FindGlobalScale._find_scale(wcs, scale_guess='bla')

    @staticmethod
    def random_wcs_with_random_reference_lines(num_lines=200, cntd_error=0.0, true_scale=1E5, ml=None):
        # note centroid error cntd_error is the error in wavelength space, not pixel space.
        m0 = np.random.randint(10, 100)
        true_scale = np.random.normal(true_scale, true_scale/10)
        pix = np.random.randint(10, 1000, size=num_lines)
        orders = np.random.randint(0, 4, size=num_lines)
        normed_pix, normed_orders = wcsu.normalize_coordinates(pix, 1000, 0), wcsu.normalize_coordinates(orders, 4, 0)
        measured_lines = {'normed_pixel': normed_pix[ml:], 'normed_order': normed_orders[ml:], 'order': orders[ml:]}
        wcs = WavelengthSolution(min_order=0, max_order=4, min_pixel=0, max_pixel=1000,
                                 measured_lines=measured_lines, model={0: [0, 1, 2], 1: [0, 1]},
                                 model_coefficients=np.random.normal(2, .3, size=5), m0=m0)
        reference_lines = wcs(pix, orders) * true_scale
        wcs.reference_lines = reference_lines + np.random.normal(0, cntd_error)  # centroiding errors, sort of.
        return wcs, true_scale

    @pytest.mark.integration
    def test_accuracy_best_case(self):
        wcs, true_scale = self.random_wcs_with_random_reference_lines(num_lines=20, cntd_error=0.0, true_scale=1E5)
        scale = FindGlobalScale._find_scale(wcs, scale_guess=1.1*true_scale, rrange=(0.1, 1.5))
        assert np.isclose(scale, true_scale)

    @pytest.mark.integration
    def test_accuracy_with_centroiding_error(self):
        err = 10
        wcs, true_scale = self.random_wcs_with_random_reference_lines(num_lines=80, cntd_error=err, true_scale=1E5)
        scale = FindGlobalScale._find_scale(wcs, scale_guess=1.1*true_scale, rrange=(0.1, 1.5))
        assert np.isclose(scale, true_scale, atol=100 * err)

    @pytest.mark.integration
    def test_accuracy_with_missed_lines(self):
        err = 10
        missed_lines = 40
        wcs, true_scale = self.random_wcs_with_random_reference_lines(num_lines=80, cntd_error=err,
                                                                      true_scale=1E5, ml=missed_lines)
        scale = FindGlobalScale._find_scale(wcs, scale_guess=1.1*true_scale, rrange=(0.1, 1.5))
        assert np.isclose(scale, true_scale, rtol=0.01)


class TestSolutionRefineInitial:
    @mock.patch('xwavecal.wavelength.SolutionRefineInitial.constrain_solution_over_detector')
    def test_do_stage_fiber(self, fake_constrain):
        image = FakeImage()
        context = FakeContext()
        context.intermediate_wavelength_model = {1: [0], 2: [0, 1]}
        image.wavelength_solution = {1: WavelengthSolution(model_coefficients=np.arange(2), model={1: [0], 2: [0]},
                                                           min_order=0, max_order=5, min_pixel=0, max_pixel=400)}
        def constrain(wcs, *args, **kwargs):
            return wcs, 1
        fake_constrain.side_effect = constrain
        image = SolutionRefineInitial(context).do_stage_fiber(image, fiber=1)
        assert image.wavelength_solution[1].model == {1: [0], 2: [0, 1]}

    def test_selecting_orders(self):
        wcs = WavelengthSolution(min_order=0, max_order=7, overlap_range=(2, 4))
        order_ranges = SolutionRefineInitial._ranges_to_evaluate(*wcs.overlap_range, wcs.min_order, wcs.max_order)
        assert np.allclose(order_ranges, [(2, 4), (1, 5), (0, 6), (-1, 7)])

    @pytest.mark.integration
    def test_accuracy_constrain_solution_over_detector(self):
        m0 = np.random.randint(10, 100)
        wcs = Utils.simple_wcs(m0)
        wcs, residuals = SolutionRefineInitial.constrain_solution_over_detector(wcs)
        assert np.isclose(np.mean(residuals), 0, atol=1E-5)

    def test_convergence_criterion(self):
        assert not SolutionRefineInitial._converged_when_max_iter()


class TestSolutionRefineFinal:
    @mock.patch('xwavecal.wavelength.SolutionRefineFinal._refine')
    def test_do_stage_fiber(self, fake_refine):
        image = FakeImage()
        context = FakeContext()
        context.final_wavelength_model = {1: [0], 2: [0, 1]}
        image.wavelength_solution = {1: WavelengthSolution(model={1: [0], 2: [0]})}

        def refine(wcs, final_model, *args, **kwargs):
            wcs.model = final_model
            return wcs, 1
        fake_refine.side_effect = refine
        image = SolutionRefineFinal(context).do_stage_fiber(image, fiber=1)
        assert image.wavelength_solution[1].model == {1: [0], 2: [0, 1]}

    @mock.patch('xwavecal.wavelength.refine_wcs')
    def test_refine(self, fake_refine):
        def refine(wcs, *args, **kwargs):
            wcs.models_tried.append(wcs.model)
            return wcs, 'residuals_passed'
        fake_refine.side_effect = refine
        wcs = WavelengthSolution(min_order=1, max_order=3, min_pixel=0, max_pixel=30,
                                 model={0: [0], 1: [0]},
                                 model_coefficients=np.array([1.0, 1.0]),
                                 reference_lines=np.array([1, 2]), m0=52, grating_eq=True)
        wcs.models_tried = []
        final_model = {0: [0, 1], 1: [0, 1]}
        wcs, residuals = SolutionRefineFinal._refine(wcs, final_model)
        assert residuals is 'residuals_passed'
        assert wcs.models_tried[0] == {0: [0, 1], 1: [0]}
        assert wcs.models_tried[1] == final_model

    def test_convergence(self):
        converged = False
        for residuals in [[1, 2, 3], [1, 2, 2], [2, 2, 2], [10, 2, 2]]:
            if SolutionRefineFinal._converged(residuals, [2, 2, 2]):
                converged = True
                break
        assert converged


class TestSolutionRefineOnce:
    @mock.patch('xwavecal.wavelength.refine_wcs')
    def test_do_stage_fiber(self, fake_refine):
        image = FakeImage()
        context = FakeContext()
        image.wavelength_solution = {1: WavelengthSolution(model={1: [0], 2: [0]})}

        def refine(wcs, *args, **kwargs):
            wcs.model = {1: [0], 2: [0, 1]}
            return wcs, 1
        fake_refine.side_effect = refine
        image = SolutionRefineOnce(context).do_stage_fiber(image, fiber=1)
        assert image.wavelength_solution[1].model == {1: [0], 2: [0, 1]}


class TestApplyToSpectrum:
    @pytest.mark.integration
    def test_do_stage(self):
        image = FakeImage()
        context = FakeContext()
        image.data_tables = {context.main_spectrum_name: Table({'ref_id': [0, 1],
                                                                     'fiber': [1, 2],
                                                                     'flux': [[0, 1, 2], [0, 1, 2]],
                                                                     'pixel': [[-1, 0, 1], [-1, 0, 1]],
                                                                     'wavelength': np.zeros((2, 3))})}
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 1, 1
        image.wavelength_solution = {0: None, 1: Utils.simple_wcs(), 2: Utils.simple_wcs()}
        image = ApplyToSpectrum(context).do_stage(image)
        spectrum = image.data_tables[context.main_spectrum_name]
        assert not np.allclose(spectrum['wavelength'], 0)
        assert len(spectrum.colnames) == 5


class TestBlazeCorrectArcEmissionLines:
    def test_fluxes_are_assigned(self):
        image = FakeImage()
        context = FakeContext()
        image.data_tables = {context.blaze_corrected_spectrum_name: Table({'ref_id': [0, 1, 0],
                                                                     'fiber': [1, 1, 0],
                                                                     'flux': [[0, 10, 2], [0, 20, 2], [0, 15, 2]],
                                                                     'pixel': [[0, 1, 2], [0, 1, 2], [0, 1, 2]]})}
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 1, 0
        image.wavelength_solution = {1: WavelengthSolution()}
        image.wavelength_solution[1].measured_lines = {'pixel': np.array([1, 1]), 'order': np.array([0, 1]),
                                                       'flux': np.array([-9, -8]),
                                                       'corrected_flux': np.array([80, 60])}
        image = BlazeCorrectArcEmissionLines(context).do_stage(image)
        assert np.allclose(image.wavelength_solution[1].measured_lines['corrected_flux'], [10, 20])

    def test_fluxes_are_assigned_on_missing_blaze_correction(self):
        image = FakeImage()
        context = FakeContext()
        image.data_tables = {}
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 1, 0
        image.wavelength_solution = {1: WavelengthSolution()}
        image.wavelength_solution[1].measured_lines = {'pixel': np.array([1, 1]), 'order': np.array([0, 1]),
                                                       'flux': np.array([-9, -8])}
        image = BlazeCorrectArcEmissionLines(context).do_stage(image)
        assert np.allclose(image.wavelength_solution[1].measured_lines['corrected_flux'],
                           image.wavelength_solution[1].measured_lines['flux'])


class TestTabulateArcEmissionLines:
    def test_format_lines(self):
        image = FakeImage()
        image.wavelength_solution = {0: WavelengthSolution(measured_lines={'pixel': [0, 1]}),
                                     1: WavelengthSolution(measured_lines={'pixel': [1, 2, 3]})}
        lines = TabulateArcEmissionLines._format_lines(image, fibers=[0, 1])
        assert all([name in lines.colnames for name in ['fiber', 'pixel', 'wavelength', 'reference_wavelength']])
        assert np.allclose(lines['pixel'], [0, 1, 1, 2, 3])
        assert np.allclose(lines['fiber'], [0, 0, 1, 1, 1])

    def test_do_stage(self):
        image = FakeImage()
        context = FakeContext()
        image.data_tables = {}
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 1, 1, 0
        image.wavelength_solution = {0: Utils.simple_wcs(50), 1: Utils.simple_wcs(52)}
        image = TabulateArcEmissionLines(context).do_stage(image)
        lines = image.data_tables[context.emission_lines_table_name]
        length = len(image.wavelength_solution[0].measured_lines['pixel']) * 2
        assert np.allclose([len(lines['fiber']), len(lines['pixel']), len(lines['wavelength'])], length)
        assert np.allclose([len(lines['reference_wavelength']), len(lines['order'])], length)


def test_normalize_coordinates():
    normed = wcsu.normalize_coordinates(np.arange(3, 6), max_value=5, min_value=3)
    assert np.allclose(normed, [-1, 0, 1])


def test_normalize_coordinates_raises_error():
    with pytest.raises(Exception):
        wcsu.normalize_coordinates(np.arange(3, 6), max_value=3, min_value=5)


class TestLineIdentification:
    @mock.patch('xwavecal.utils.wavelength_utils.find_peaks',
                return_value=(np.array([0.1, 1.1]), np.array([0.01, 0.02]), np.array([0, 1])))
    def test_identify_lines(self, fake_peaks):
        spectrum = Table({'flux': [[10, 20]], 'pixel': [[1, 2]], 'id': [32]})
        lines = wcsu.identify_lines(spectrum, stderr=10, order_key='id')
        assert np.allclose(lines['flux'], [10, 20])
        assert np.allclose(lines['order'], [32, 32])
        assert np.allclose(lines['pixel'], [0.1, 1.1])
        assert np.allclose(lines['pixel_err'], [0.01, 0.02])


class TestRefineUtils:
    def test_restrict(self):
        lines = {'order': np.arange(10), 'misc': np.arange(10, 20)}
        r_lines = wcsu.restrict(lines, 2, high_order=8)
        assert np.allclose(r_lines['order'], np.arange(2, 9))
        assert np.allclose(r_lines['misc'], np.arange(12, 19))

    def test_restrict_returns_original(self):
        lines = {'order': np.arange(10), 'misc': np.arange(10, 20)}
        r_lines = wcsu.restrict(lines)
        assert np.allclose(r_lines['order'], lines['order'])
        assert np.allclose(r_lines['misc'], lines['misc'])

    @staticmethod
    def no_clip(wcs, measured_lines, reference_lines, **kwargs):
        return measured_lines

    @staticmethod
    def converge(residuals, *args, **kwargs):
        return True

    def test_refine_wcs_accuracy(self):
        # this fails very rarely, but the occasional failure is annoying.
        # So we run the test twice and look for at least one success.
        success = []
        for i in range(2):
            m0 = np.random.randint(10, 100)
            wcs = Utils.simple_wcs(m0)
            wcs, residuals = refine_wcs(wcs, wcs.measured_lines, wcs.reference_lines, self.converge, self.no_clip)
            success.append(np.isclose(np.mean(residuals), 0, atol=1E-5))
        assert any(success)

    def test_clip_returns_lines_without_residuals(self):
        assert np.allclose([1, 2, 3], wcsu._sigma_clip([], [1, 2, 3]))

    def test_clip(self):
        clipped_lines = wcsu._sigma_clip([10, 10, *np.ones(30)], {'a': np.arange(32)}, sigma=3, stdfunc=np.std)
        assert np.allclose(clipped_lines['a'], np.arange(32)[2:])


class TestIdentifyPrincipleOrderNumber:
    CONTEXT = FakeContext()

    @mock.patch('xwavecal.wavelength.IdentifyPrincipleOrderNumber.merit_per_m0', return_value=([1, .09, 1], np.array([2, 3, 4])))
    def test_do_stage_fiber(self, mock_merit):
        image = FakeImage()
        image.wavelength_solution = {1: Utils.simple_wcs(m0=1000)}
        image = IdentifyPrincipleOrderNumber(self.CONTEXT).do_stage_fiber(image, 1)
        assert image.wavelength_solution[1].m0 == 3

    @mock.patch('xwavecal.wavelength.IdentifyPrincipleOrderNumber.merit_per_m0', return_value=([1, .5, 1], np.array([2, 3, 4])))
    def test_failure(self, mock_merit):
        image = FakeImage()
        image.wavelength_solution = {1: Utils.simple_wcs(m0=1000)}
        image = IdentifyPrincipleOrderNumber(self.CONTEXT).do_stage_fiber(image, 1)
        assert image.wavelength_solution[1] is None

    def test_merit_per_m0(self):
        image = FakeImage()
        image.wavelength_solution = {1: Utils.simple_wcs(m0=10)}
        identifier = IdentifyPrincipleOrderNumber(self.CONTEXT)
        identifier.STAGES_TODO = []
        merit, m0_vals = identifier.merit_per_m0(image, 1, (5, 11, 5)) # try m0=5 and 10.
        assert merit[1] < merit[0]
        assert np.allclose(m0_vals, [5, 10])


class TestOnSyntheticData:
    @pytest.mark.integration
    def test_find_feature_wavelengths(self):
        # Generate emission line positions from a real wavelength solution on NRES.
        num_orders = 67
        wcs = WavelengthSolution(model={0: [0, 1, 2, 3, 4, 5],
                                        1: [0, 1, 2, 3, 4, 5],
                                        2: [0, 1, 2, 3, 4, 5],
                                        3: [0, 1, 2, 3, 4, 5],
                                        4: [0]},
                                 min_order=0, max_order=num_orders - 1,
                                 min_pixel=0, max_pixel=4095, m0=52,
                                 model_coefficients=np.array([4.66146591e+05, -7.74683540e+02, -3.54722161e+02, -8.99350361e+01,
                                                              -1.70864902e+01, -2.49239167e+00,  4.67592938e+03,  1.23351661e+02,
                                                              6.16440684e+01,  2.03181960e+01,  3.20985475e+00,  1.26399172e-02,
                                                              -5.20952098e+02, -2.52553564e+01, -1.41207709e+01, -1.80775282e+00,
                                                              -8.57769848e-01,  1.75480247e+00,  1.95194420e+01,  3.81273242e+00,
                                                              7.61239901e-01, -8.78661806e-01, -1.73232452e+00, -1.96449284e+00,
                                                              -3.57462413e+00]))

        measured_lines, line_list = SpectrumUtils().generate_measured_lines(n_list=1000, n_true=0,
                                                                            n_garb=0, wcs=wcs)
        measured_lines['fiber'] = np.ones_like(measured_lines['order'])
        wavelength_models = {'initial_wavelength_model': {1: [0, 1, 2], 2: [0, 1, 2]},
                             'intermediate_wavelength_model': {0: [0, 1, 2], 1: [0, 1, 2], 2: [0, 1, 2]},
                             'final_wavelength_model': copy.copy(wcs.model)}
        measured_lines['wavelength'] = find_feature_wavelengths(measured_lines, line_list,
                                                                m0_range=(51, 54), max_pixel=4096, min_pixel=0,
                                                                wavelength_models=wavelength_models)
        assert np.allclose(measured_lines['wavelength'], measured_lines['true_wavelength'], rtol=1e-8)


class Utils:
    @staticmethod
    def simple_wcs(m0=10):
        line_orders, line_normed_pixels = np.array([1., 1, 2, 2, 3, 3]), np.array([-1., 1, -1, 1, -1, 1])
        measured_lines = {'normed_pixel': line_normed_pixels, 'normed_order': np.array([-1., -1, 0, 0, 1, 1]),
                          'order': line_orders, 'pixel': line_normed_pixels}
        reference_lines = 1./(m0 + line_orders) * (1. + line_normed_pixels)
        wcs = WavelengthSolution(min_order=1, max_order=3, overlap_range=(2, 3), measured_lines=measured_lines,
                                 model={0: [0], 1: [0]}, model_coefficients=np.array([1.01, 1.02]),
                                 reference_lines=reference_lines, m0=m0, grating_eq=True)
        return wcs



import mock
from astropy.io.fits import HDUList, PrimaryHDU
from tempfile import TemporaryDirectory

from xwavecal.utils.fits_utils import writeto
from xwavecal.utils.instrument_specific import parse_nres_region_keyword, parse_harps_region_keyword


class TmpDir(TemporaryDirectory):
    def __init__(self, suffix=None, prefix=None, dir=None):
        super(TmpDir, self).__init__(suffix=suffix, prefix=prefix, dir=dir)

    def __enter__(self):
        return 't3mp'


class TestWriteTo:
    @mock.patch('astropy.io.fits.HDUList.writeto')
    def test_writeto_no_fpack(self, mock_write):
        hdu_list = HDUList([PrimaryHDU()])
        writeto(hdu_list, 'path.fz', fpack=False, overwrite=True, output_verify='verify')
        mock_write.assert_called_with('path', overwrite=True, output_verify='verify')

    @mock.patch('shutil.move')
    @mock.patch('os.system')
    @mock.patch('astropy.io.fits.HDUList.writeto')
    @mock.patch('tempfile.TemporaryDirectory', new=TmpDir)
    def test_writeto_fpack(self, mock_write, mock_sys, mock_move):
        hdu_list = HDUList([PrimaryHDU()])
        writeto(hdu_list, 'path', fpack=True, overwrite=True, output_verify='verify', quant=64)
        mock_write.assert_called_with('t3mp/path', overwrite=True, output_verify='verify')
        mock_sys.assert_called_with('fpack -q 64 t3mp/path')
        mock_move.assert_called_with('t3mp/path.fz', 'path.fz')


def test_parse_region_keyword():
    assert (slice(0, 4096), slice(0, 4096)) == parse_nres_region_keyword('[1:4096,1:4096]')


def test_parse_harps_region_keyword():
    key = (4096, 4096, 50, 0, 50, 0, 4196, 4096)
    assert (slice(0, 4096), slice(50, 4146)) == parse_harps_region_keyword(key)
    key = (50, 0, 50, 0, 50, 0, 4196, 4096)
    assert (slice(0, 4096), slice(4146, 4196)) == parse_harps_region_keyword(key)



import tempfile
import os
from datetime import datetime, timedelta
import sqlite3
import mock

import xwavecal.database as db
from xwavecal.tests.utils import FakeImage


def test_format_db_info():
    data = FakeImage()
    fmt = '%Y-%m-%dT%H:%M:%S.%f'
    db_info = db.format_db_info(data, fmt)
    fib0, fib1, fib2 = db_info['fiber0'], db_info['fiber1'], db_info['fiber2']
    assert all([db_info['type'] == data.get_header_val('type'),
                db_info['observation_date'] == datetime.strptime(data.get_header_val('observation_date'), fmt),
                db_info['date_created'] - datetime.now() <= timedelta(seconds=20),
                db_info['instrument'] == data.get_header_val('instrument'),
                db_info['instrument2'] == data.get_header_val('instrument2'),
                db_info['site_name'] == data.get_header_val('site_name'),
                fib0 == data.fiber0_lit, fib1 == data.fiber1_lit, fib2 == data.fiber2_lit,
                db_info['is_bad'] == 0, db_info['filepath'] == data.filepath])


@mock.patch('xwavecal.database.query_db_for_match', return_value=None)
def test_add_to_db(mock_match):
    data = FakeImage()
    fmt = '%Y-%m-%dT%H:%M:%S.%f'
    db_info = db.format_db_info(data, fmt)
    with tempfile.TemporaryDirectory() as temp_directory:
        db_path = os.path.join(temp_directory, 'test.db')
        db.add_data_to_db(db_path, db_info)
        db.add_data_to_db(db_path, db_info)
        db.add_data_to_db(db_path, db_info)
        # expect 3 identical entries in the database.
        conn = sqlite3.connect(db_path)
        c = conn.cursor()
        c.execute('SELECT * FROM caldata')
        assert len(c.fetchall()) == 3
        conn.close()


def test_update_db():
    data = FakeImage()
    fmt = '%Y-%m-%dT%H:%M:%S.%f'
    db_info = db.format_db_info(data, fmt)
    with tempfile.TemporaryDirectory() as temp_directory:
        db_path = os.path.join(temp_directory, 'test.db')
        db.add_data_to_db(db_path, db_info)
        db.add_data_to_db(db_path, db_info)
        db.add_data_to_db(db_path, db_info)
        # expect the 3 identical entries in the database to be culled into one.
        conn = sqlite3.connect(db_path)
        c = conn.cursor()
        c.execute('SELECT * FROM caldata')
        assert len(c.fetchall()) == 1
        conn.close()


def test_query_database():
    data1, data2, data3 = FakeImage(), FakeImage(), FakeImage()
    reference_data = FakeImage()
    data3.set_header_val('type', 'wrong')
    data3.set_header_val('observation_date', '2019-04-15T12:56:44.466')
    data2.set_header_val('observation_date', '2019-04-13T12:56:44.466')
    data2.filepath = 'some/path'
    data1.set_header_val('observation_date', '2019-04-11T12:56:44.466')
    reference_data.set_header_val('observation_date', '2019-04-16T12:56:44.466')
    fmt = '%Y-%m-%dT%H:%M:%S.%f'
    with tempfile.TemporaryDirectory() as temp_directory:
        db_path = os.path.join(temp_directory, 'test.db')
        for data in [data1, data2, data3]:
            db_info = db.format_db_info(data, fmt)
            db.add_data_to_db(db_path, db_info)
        filepath = db.query_db_for_nearest(db_path, reference_data, 'lampflat', fmt)
        assert filepath == data2.filepath
        assert db.query_db_for_nearest(db_path, reference_data, 'no_type', fmt) is None


def test_query_database_returns_none():
    assert db.query_db_for_nearest('non/existent/path', None, 'lampflat', '%Y-%m-%dT%H:%M:%S.%f') is None
    assert db.query_db_for_nearest('non/existent/path', None, 'lampflat', '%Y-%m-%dT%H:%M:%S.%f') is None



import numpy as np

from scipy import signal, optimize, interpolate
import warnings


def normalize_by_brightest(two_d_spectrum, n=10):
    """
    :param two_d_spectrum: ndarray 2 dimensional, where two_d_spectrum[i] is the ith one d spectrum.
    :return: two_d_spectrum where each row has been divided by norm,
             where norm is the median of the flux from the brightest n pixels of that row.
    NOTE: norm can be no smaller than 1.
    """
    temp = np.copy(two_d_spectrum)
    n = min(two_d_spectrum.shape[1], n)
    for i in range(two_d_spectrum.shape[0]):
        brightest_pixels = np.sort(temp[i])[::-1][:n]
        temp[i] /= max(np.median(brightest_pixels), 1)
    return temp


def n_largest_elements(arr, n):
    """
    :param arr: ndarray
    :param n: int
    :return: returns the indices for the n largest elements of arr.
    """
    return arr.argsort()[-n:][::-1]


def minmax(arr):
    return np.min(arr), np.max(arr)


def overlap_region(a, b):
    """
    :param a: ndarray or list
    :param b: ndarray or list
    :return: (left, right):
             left: int. smallest value of the intersection of a and b if they were continuous arrays.
             right: int. largest value of the intersection of a and b if they were continuous arrays.
             returns None if there is no overlap
    Examples:
            overlap_region([1, 2.1], [1.9, 3]) returns (1.9, 2.1)
            overlap_region([1, 2.1], [2.2, 3]) returns None
    """
    left, right = np.max([np.min(a), np.min(b)]), np.min([np.max(a), np.max(b)])
    return (left, right) if right > left else None


def find_nearest(array_a, array_b):
    """
    # TODO consider replacing with np.searchsorted.
    :param array_a: ndarray
    :param array_b: ndarray. Must be Sorted.
    :return: a list of elements of B which are closest to each element in A, searched for in order of A.
    e.g. if A = [1, 2, 2, 2] and B = [0.99, 1.99, 4]
    then we would return [0.99, 1.99, 1.99, 1.99]
    Believe it or not, this is 1000 times faster (for a list of length 10000) then a python implementation
    of bisection-searching array_b for every element in array_a.
    """
    find_func = interpolate.interp1d(array_b, array_b, kind='nearest', bounds_error=False,
                                     fill_value=(np.min(array_b), np.max(array_b)))
    return find_func(array_a)


def find_peaks(y, x, yerr, height=0, distance=1, prominence=None, window=6, peak_width=2):
    """
    :param y: ndarray. float. Amplitude as a function of x
    :param x: ndarray. dtype=int. x coordinates for y.
    :param yerr: ndarray. float. 1 sigma errors for y.
    :param height: float. peaks with amplitudes below height will be excluded. scalar or ndarray of length y.
    :param distance: int. minimum number of grid points (distance) between two identified peaks.
    :param prominence: float. The vertical distance between the peak and its lowest contour line.
                       Recommend prominence = 0.5 * np.abs(y)
    :param peak_width: estimate for the full width at half max of the peaks.
    :param window: int. A guassian will be fit around the peak but within a window of width.
    :return: ndarray, ndarray, ndarray
             peaks, peak_errs, peak_indices
             peaks is the peak centroid in the given x coordinates
             peak_errs is the 1 sigma error on each centroid in peaks.
             peak_indices the indices of the closest x coordinate to each location in peak_locations.

    Note on prominence:
        Strategy to compute a peak’s prominence:
        - Extend a horizontal line from the current peak to the left and right until the line either
          reaches the window border (see wlen) or intersects the signal again at the slope of a
          higher peak. An intersection with a peak of the same height is ignored.
        - On each side find the minimal signal value within the interval defined above.
          These points are the peak’s bases.
        - The higher one of the two bases marks the peak’s lowest contour line. The prominence can
          then be calculated as the vertical difference between the peaks height itself and
          its lowest contour line.

    Note: might do better to clip y for all points below height, instead of using signal.find_peaks height argument.
    Note: If the gaussian fit a peak position which differs by more than 3 from the initial guess,
          then we return the initial guess.
    """
    peak_indices = signal.find_peaks(y, distance=distance, prominence=prominence, height=height)[0]  # identify peaks.
    peaks = fit_peaks(x, y, yerr, window // 2, peak_indices, std=peak_width)  # fit a Gaussian to each peak.
    # remove fits that were wildly different from init guess, or have non finite errors:
    bad_fits = np.where(np.logical_or(~np.isclose(peaks[:, 0], x[peak_indices], atol=5), ~np.isfinite(peaks[:, 1])))
    peaks[bad_fits, 0], peaks[bad_fits, 1] = x[peak_indices[bad_fits]], 1
    return peaks[:, 0].flatten(), peaks[:, 1].flatten(), peak_indices


def fit_peaks(x, y, yerr, half_width, index_guess, std):
    peaks = []
    amplitude_guess = y[index_guess]
    for i, ampl in zip(index_guess, amplitude_guess):
        sl = slice(i - half_width, i + half_width + 1)
        center, err = x[i], 1
        if len(x[sl]) >= 3:  # num params in the model
            try:
                params, pcov = optimize.curve_fit(gaussian, x[sl], y[sl], [ampl, center, std],
                                                  sigma=yerr[sl], absolute_sigma=True)
                center, err = params[1], np.sqrt(np.diag(pcov))[1]
            except RuntimeError as e:
                warnings.warn(str(e))

        peaks.append([center, err])
    return np.array(peaks)


def gaussian(x, ampl, mu, std):
    return ampl * np.exp(-np.power(x - mu, 2.) / (2 * np.power(std, 2.)))


def brute_local_min(fun, x0, args, rrange=(0.1, 10), filtw=501, step=10, finish='Nelder-Mead', **kwargs):
    """
    :param fun: function you wish to find the deepest local minimum for. fun(x, args) must be able to
                accept an ndarray x and return an ndarray of fun evaluated at all the points in x.
    :param x0: initial guess
    :param args: extra args to pass to fun
    :param rrange: (low, high). The brute force minimization over the grid low*x0 and high*x0 with spacing step.
    :param filtw: the filter window for the median filter of fun() evaluating over the grid of points
    :param step: interval size in the brute force grid
    :param finish: Finish method to pass to scipy.optimize.minimize to refine the brute force minimum.
                   Note: use a method like Newton-Raphson that prefers local minima.
    :param kwargs: dict. Placeholder which allows scipy.optimize to pass standard extraneous arguments.
    :return: OptimizeResult with OptimizeResult.x = np.array([minimum]).
    Note: this is intended to be used with optimize.minimize with optimize.minimize(..., method=brute_local_min)
    """
    eval_points = np.arange(x0 * rrange[0], x0 * rrange[1], step=step)
    fun_vals = np.array(fun(eval_points, *args))
    fun_vals /= signal.medfilt(fun_vals, filtw)
    minimum_pt = eval_points[fun_vals == np.min(fun_vals[filtw:-filtw])][:1]
    if finish is not None:
        minimum_pt = optimize.minimize(fun, x0=minimum_pt, args=args, method=finish).x
    return optimize.OptimizeResult(x=minimum_pt)
